@@ -1,7 +1,7 @@
 /* pam_ftp module */
 
 /*
- * $Id: pam_ftp.c,v 1.11 2007/03/29 13:45:38 t8m Exp $
+ * $Id$
  *
  * Written by Andrew Morgan <morgan@linux.kernel.org> 1996/3/11
  *
@@ -79,7 +79,7 @@ static int lookup(const char *name, const char *list, const char **_user)
     if (list && *list) {
 	const char *l;
 	char *list_copy, *x;
-	char *sptr;
+	char *sptr = NULL;
 
 	list_copy = x_strdup(list);
 	x = list_copy;
@@ -162,6 +162,7 @@ pam_sm_authenticate (pam_handle_t *pamh, int flags UNUSED,
 			       GUEST_LOGIN_PROMPT);
 
 	if (retval != PAM_SUCCESS) {
+	    _pam_overwrite (resp);
 	    _pam_drop (resp);
 	    return ((retval == PAM_CONV_AGAIN)
 		    ? PAM_INCOMPLETE:PAM_AUTHINFO_UNAVAIL);
@@ -171,7 +172,7 @@ pam_sm_authenticate (pam_handle_t *pamh, int flags UNUSED,
 	  /* XXX: Some effort should be made to verify this email address! */
 
 	    if (!(ctrl & PAM_IGNORE_EMAIL)) {
-		char *sptr;
+		char *sptr = NULL;
 		token = strtok_r(resp, "@", &sptr);
 		retval = pam_set_item(pamh, PAM_RUSER, token);
 
@@ -200,6 +201,7 @@ pam_sm_authenticate (pam_handle_t *pamh, int flags UNUSED,
 	}
 
 	/* clean up */
+	_pam_overwrite(resp);
 	_pam_drop(resp);
 
 	/* success or failure */
